{ ****************************************************************** }
{                                                                    }
{   VCL component TGPopulation                                       }
{                                                                    }
{   Genetic Programming Population                                   }
{                                                                    }
{   Code generated by Component Create for Delphi                    }
{                                                                    }
{   Generated from source file e:\code\gp\gpopulation\gpunit4.cd     }
{   on 25 July 2001 at 1:17                                          }
{                                                                    }
{   Copyright  1999 by Adi Andrei  aandrei@writeme.com              }
{                                                                    }
{ ****************************************************************** }

unit GPUnit;

interface

uses WinTypes, WinProcs, Messages, SysUtils, Classes, Controls, 
     Forms, Dialogs;

{constants, can be changed to define the maximum capabilities of the component}

const
  pop_max = 5000; {maximum # of individuals in a population }
const
  term_max = 100; {maximum # of terminals in t_set }
const
  fn_max = 100; {maximum # of functions in a fn_set }

const
  ind_len = 65535; {maximum length of an individual}

type
  TIndividual = class(TObject)
  public
    fitness: integer;
    body: string; {array[1..ind_len] of char}
    constructor Create;
    function fit: integer;
    procedure init;
    function eval: integer;
    procedure setfit(f: integer);
  end;



Type
  EGeneticError = class(Exception);

type
 rulete_array = array[0..pop_max-1] of longint;

type
 ind_array = array[0..pop_max-1] of Tindividual;

type
  percent = 0..100;

type
  TTerminal = string[10];

type
  TFunction = record
    fname: string[10];{function name}
    n_par: integer;{number of parameters}
  end;


type
 Term_Array = array[0..term_max-1] of TTerminal;

type
 Fn_Array = array[0..fn_max-1] of TFunction;






type
  TIndEvalEvent = procedure (Sender : TObject; Body:string) of object;

  TGPopulation = class(TComponent)
    private
      { Private fields of TGPopulation }
        { Storage for property BestIndex }
        FBestIndex : Integer;
        { Storage for property CrossoverP }
        FCrossoverP : Percent;
        { Storage for property Elite }
        FElite : Boolean;
        { Storage for property FitStop }
        FFitStop : Integer;
        { Storage for property GenCount }
        FGenCount : Integer;
        { Storage for property IndCount }
        FIndCount : Integer;
        { Storage for property IndLength }
        FIndLength : Integer;
        { Storage for property InstantFitness }
        FInstantFitness : Integer;
        { Storage for property ZeroFnP }
        FZeroFnP : Percent;
        { Storage for property ZeroIndLenP }
        FZeroIndLenP : Percent;
        { Pointer to application's OnAfterIndEval handler, if any }
        FOnAfterIndEval : TNotifyEvent;
        { Pointer to application's OnAfterPopEval handler, if any }
        FOnAfterPopEval : TNotifyEvent;
        { Pointer to application's OnBeforeIndEval handler, if any }
        FOnBeforeIndEval : TNotifyEvent;
        { Pointer to application's OnBeforePopEval handler, if any }
        FOnBeforePopEval : TNotifyEvent;
        { Pointer to application's OnFitReached handler, if any }
        FOnFitReached : TNotifyEvent;
        { Pointer to application's OnIndEval handler, if any }
        FOnIndEval : TIndEvalEvent;

      { Private methods of TGPopulation }
        { Method to set variable and property values and create objects }
        procedure AutoInitialize;
        { Method to free any objects created by AutoInitialize }
        procedure AutoDestroy;

    protected
      { Protected fields of TGPopulation }
        { No of functions in the function set. }
        FnCount : Integer;
        { Mating pool. }
        Ind2 : Ind_array;
        { Selection roulete }
        Roulete : Rulete_array;
        { No of terminals in the terminal symbols set. }
        TermCount : Integer;
        { maneuver individual }
        X : TIndividual;

      { Protected methods of TGPopulation }
        { Method to generate OnAfterIndEval event }
        procedure AfterIndEval(Sender : TObject); virtual;
        { Method to generate OnAfterPopEval event }
        procedure AfterPopEval(Sender : TObject); virtual;
        { Method to generate OnBeforeIndEval event }
        procedure BeforeIndEval(Sender : TObject); virtual;
        { Method to generate OnBeforePopEval event }
        procedure BeforePopEval(Sender : TObject); virtual;
        { Method to generate OnFitReached event }
        procedure FitReached(Sender : TObject); virtual;
        { Method to generate OnIndEval event }
        procedure IndEval(Sender : TObject; Body:string); virtual;
        { Returns the index of the best individual. }
        function Best:integer;
        { Generate an error. }
        procedure Err(m:string);
        procedure GenX;
        { Initialize the roulete }
        procedure InitRoulete;
        procedure Loaded; override;
        { Perform crossover on to s-expressions }
        function cross(s1:string;s2:string):string;
        function draw:integer;

    public
      { Public fields and properties of TGPopulation }
        { Function set. }
        FnSet : Fn_Array;
        { Generation running. }
        Gen : Integer;
        { Original population. }
        Ind : Ind_array;
        { Is the 'execute' method running ? }
        Running : Boolean;
        { Terminal symbols set. }
        TermSet : Term_array;

      { Public methods of TGPopulation }
        { Add function to the function set. }
        procedure AddFn(f: string; p:integer);
        { Add a terminal to the terminal symbols set. }
        procedure AddTerm(t:string);
        { Constructor }
        constructor Create(AOwner: TComponent); override;
        { Destructor }
        destructor Destroy; override;
        { Evaluate Population. }
        procedure Eval;
        function Execute : Boolean;
        { Run Generation Zero. }
        procedure GenZero;
        { Returns the number of functions in the function set }
        function GetFnCount:integer;
        { Returns the number of terminals in the terminal set }
        function GetTermCount : integer;
        function GoON:boolean;
        { Initialize population. Reset the terminal / function set }
        procedure Init;
        { Run (evolve) the next generation. }
        procedure NextGen;
        { Stop running started by Execute method }
        procedure Stop;

    published
      { Published properties of TGPopulation }
        property OnAfterIndEval : TNotifyEvent read FOnAfterIndEval write FOnAfterIndEval;
        property OnAfterPopEval : TNotifyEvent read FOnAfterPopEval write FOnAfterPopEval;
        property OnBeforeIndEval : TNotifyEvent read FOnBeforeIndEval write FOnBeforeIndEval;
        property OnBeforePopEval : TNotifyEvent read FOnBeforePopEval write FOnBeforePopEval;
        { Minimum specified fitness reached }
        property OnFitReached : TNotifyEvent read FOnFitReached write FOnFitReached;
        { custom individual evaluation function to be designed by the user }
        property OnIndEval : TIndEvalEvent read FOnIndEval write FOnIndEval;
        { Index of the best individual in the population. }
        property BestIndex : Integer
             read FBestIndex write FBestIndex
             default 1;
        { percent of individuals to perform crossover on }
        property CrossoverP : Percent
             read FCrossoverP write FCrossoverP
             default 50;
        { Keep the best individual in the next generation }
        property Elite : Boolean read FElite write FElite default True;
        { Fitness Stop Condition. }
        property FitStop : Integer read FFitStop write FFitStop default 90;
        { No of generations to run when using Execute method. }
        property GenCount : Integer read FGenCount write FGenCount default 50;
        { No of individuals. }
        property IndCount : Integer read FIndCount write FIndCount default 100;
        { Maximum length for an individual. }
        property IndLength : Integer
             read FIndLength write FIndLength
             default 1024;
        { Used with OnIndEval event. }
        property InstantFitness : Integer
             read FInstantFitness write FInstantFitness
             default 0;
        { Probability of functions in generation zero. }
        property ZeroFnP : Percent read FZeroFnP write FZeroFnP default 50;
        { Max length of an individual in gen zero, as percent of IndLength. }
        property ZeroIndLenP : Percent
             read FZeroIndLenP write FZeroIndLenP
             default 50;

  end;

procedure Register;

implementation


function TIndividual.eval;
begin
 { fitness := fitness_function(body);
  eval := fitness; }
  eval:=0;
end;



constructor TIndividual.create;
begin
  inherited Create;
  fitness := 0;
end;

procedure TIndividual.init;
begin
  body := '';
  fitness := 100;
end;

function TIndividual.fit;
begin
  fit := fitness;
end;

procedure TIndividual.setfit(f: integer);
begin
  fitness := f;
end;

procedure Register;
begin
     { Register TGPopulation with Genetic Programing as its
       default page on the Delphi component palette }
     RegisterComponents('Genetic Programing', [TGPopulation]);
end;

{ Method to set variable and property values and create objects }
procedure TGPopulation.AutoInitialize;
begin
     FnCount := 0;
     Gen := 0;
     Running := False;
     TermCount := 0;
     FBestIndex := 1;
     FCrossoverP := 50;
     FElite := True;
     FFitStop := 90;
     FGenCount := 50;
     FIndCount := 100;
     FIndLength := 1024;
     FInstantFitness := 0;
     FZeroFnP := 50;
     FZeroIndLenP := 50;
end; { of AutoInitialize }

{ Method to free any objects created by AutoInitialize }
procedure TGPopulation.AutoDestroy;
begin
     { No objects from AutoInitialize to free }
end; { of AutoDestroy }

{ Method to generate OnAfterIndEval event }
procedure TGPopulation.AfterIndEval(Sender : TObject);
begin
     { Has the application assigned a method to the event, whether
       via the Object Inspector or a run-time assignment?  If so,
       execute that method }
     if Assigned(FOnAfterIndEval) then
        FOnAfterIndEval(Sender);
end;

{ Method to generate OnAfterPopEval event }
procedure TGPopulation.AfterPopEval(Sender : TObject);
begin
     { Has the application assigned a method to the event, whether
       via the Object Inspector or a run-time assignment?  If so,
       execute that method }
     if Assigned(FOnAfterPopEval) then
        FOnAfterPopEval(Sender);
end;

{ Method to generate OnBeforeIndEval event }
procedure TGPopulation.BeforeIndEval(Sender : TObject);
begin
     { Has the application assigned a method to the event, whether
       via the Object Inspector or a run-time assignment?  If so,
       execute that method }
     if Assigned(FOnBeforeIndEval) then
        FOnBeforeIndEval(Sender);
end;

{ Method to generate OnBeforePopEval event }
procedure TGPopulation.BeforePopEval(Sender : TObject);
begin
     { Has the application assigned a method to the event, whether
       via the Object Inspector or a run-time assignment?  If so,
       execute that method }
     if Assigned(FOnBeforePopEval) then
        FOnBeforePopEval(Sender);
end;

{ Method to generate OnFitReached event }
procedure TGPopulation.FitReached(Sender : TObject);
begin
     { Has the application assigned a method to the event, whether
       via the Object Inspector or a run-time assignment?  If so,
       execute that method }
     if Assigned(FOnFitReached) then
        FOnFitReached(Sender);
end;

{ Method to generate OnIndEval event }
procedure TGPopulation.IndEval(Sender : TObject; Body:string);
begin
     { Has the application assigned a method to the event, whether
       via the Object Inspector or a run-time assignment?  If so,
       execute that method }
     if Assigned(FOnIndEval) then
        FOnIndEval(Sender, Body);
end;

{ Add function to the function set. }
procedure TGPopulation.AddFn(f: string; p:integer);
begin
  if FnCount < fn_max then
  begin
    FnCount:= FnCount + 1;
    FnSet[FnCount].fname := f;
    FnSet[FnCount].n_par := p;
  end
  else
  raise EGeneticError.Create(' Number of functions exceding maximum allowed');
end;

{ Add a terminal to the terminal symbols set. }
procedure TGPopulation.AddTerm(t:string);
begin
  if TermCount < term_max then
  begin
    TermCount := TermCount + 1;
    TermSet[TermCount] := t;
  end
  else raise EGeneticError.Create(' Number of terminals exceding maximum allowed');
end;

{ Returns the index of the best individual. }
function TGPopulation.Best:integer;
var
  i, m, max: integer;
begin
  m := 0; max := 0;
  for i := 0 to IndCount-1 do
  begin
    if ind[i].fit > max then
    begin
      m := i;
      max := ind[i].fit;
    end;
  end;
  Best := m;
  BestIndex:=m;
end;

{ Constructor }
constructor TGPopulation.Create(AOwner: TComponent);
var i:integer;
begin
     { Call the Create method of the parent class }
     inherited Create(AOwner);
     { AutoInitialize sets the initial values of variables and      }
     { properties; also, it creates objects for properties of       }
     { standard Delphi object types (e.g., TFont, TTimer,           }
     { TPicture) and for any variables marked as objects.           }
     { AutoInitialize method is generated by Component Create.      }
     AutoInitialize;

  x := TIndividual.create; {create the individual populations}
  for i := 0 to pop_max-1 do
  begin
    ind[i] := TIndividual.create;
    ind2[i] := TIndividual.create;
  end;

end;

{ Destructor }
destructor TGPopulation.Destroy;
var
  i: integer;
begin
     { AutoDestroy, which is generated by Component Create, frees any   }
     { objects created by AutoInitialize.                               }
     AutoDestroy;

     { Here, free any other dynamic objects that the component methods  }
     { created but have not yet freed.  Also perform any other clean-up }
     { operations needed before the component is destroyed.             }

for i := 0 to pop_max-1 do
  begin
    ind[i].free;
    ind2[i].free;
  end;
  x.free;
     { Last, free the component by calling the Destroy method of the    }
     { parent class.                                                    }
     inherited Destroy;
end;

{ Generate an error. }
procedure TGPopulation.Err(m:string);
begin
 raise EGeneticError.Create(m);
end;

{ Evaluate Population. }
procedure TGPopulation.Eval;
var
  i, f, e: integer;
begin
  if gen = 0 then e := IndCount {at gen 0 evaluate all,then evaluate only
                                the modified ones}
  else e := ((IndCount * CrossoverP) div 100);
  BeforePopEval(Self);
  for i := 0 to e-1 do
  begin
    BeforeIndEval(Self);
    {f := ind[i].eval;}
    IndEval(Self,ind[i].body);
    ind[i].fitness:=InstantFitness;
    AfterIndEval(Self);
  end;
  Best; {calculate best individual, set the bestindex value}
  AfterPopEval(Self);
end;

function TGPopulation.Execute : Boolean;
begin
     { Perform the component operation }
 Running:=true;
 Gen:=0;
 genzero;
 eval;
 Result:=GoOn;
end;

procedure TGPopulation.GenX;
var
  k, i: integer;
begin
  if (length(x.body) < (IndLength * ZeroIndLenP /100))
    and (random < ZeroFnP / 100) then { max len and  fn occuring probability}
  begin
    k := round(random(FnCount) + 1);
    x.body := x.body + '(';
    x.body := x.body + FnSet[k].fname + ' ';
    for i := 1 to FnSet[k].n_par do
      genx;
    x.body := x.body + ')';
  end
  else
  begin
    k := round(random(TermCount) + 1);
    if x.body = '' then
      x.body := '(' + TermSet[k] + ')'
    else
      x.body := x.body + TermSet[k] + ' ';
  end;
end;

{ Run Generation Zero. }
procedure TGPopulation.GenZero;
var
  i: integer;
begin
  for i := 0 to IndCount-1 do
  begin
    gen := 0;
    x.body := '';
    genx();
    ind[i].body := x.body;
  end;
end;

{ Returns the number of functions in the function set }
function TGPopulation.GetFnCount:integer;
     { Internal declarations for method }
     { type }
     { . . . }
     { var }
     { . . . }
begin
 result:=FnCount;
end;

{ Returns the number of terminals in the terminal set }
function TGPopulation.GetTermCount : integer;
     { Internal declarations for method }
     { type }
     { . . . }
     { var }
     { . . . }
begin
 result:=TermCount;
end;

function TGPopulation.GoON:boolean;
begin
     { Perform the component operation }
try
 Running:=true;
 while (
 (Gen<GenCount) and Running )do
  begin
   NextGen;
   Eval;
  if Ind[BestIndex].fitness>=FitStop then FitReached(self);
  end;
  Running:=false;
  Result := True
 except
  Running:=false;
       { Return True if the operation was successful, False otherwise }
  Result := false;
 { raise EGeneticError.Create(' Number of functions exceding maximum allowed');}
  end;
end;

{ Initialize population. Reset the terminal / function set }
procedure TGPopulation.Init;
var
  i: integer;
begin
  TermCount := 0; { # of terminals }
  FnCount := 0; { # of funtions }
  for i := 0 to IndCount-1 do
  begin
    ind[i].init;
    ind2[i].init;
  end;
end;

{ Initialize the roulete }
procedure TGPopulation.InitRoulete;
var
  i: integer;
begin
  roulete[0] := ind[0].fitness;
  for i := 1 to IndCount-1 do
  begin
    roulete[i] := roulete[i - 1] + ind[i].fitness;
  end;
end;

procedure TGPopulation.Loaded;
begin
     inherited Loaded;

     { Perform any component setup that depends on the property
       values having been set }

end;

{ Run (evolve) the next generation. }
procedure TGPopulation.NextGen;
var
  i, j: integer;
begin
{********** you must run eval before running next_gen *********}
  gen := gen + 1;
  initroulete;
  for i := 0 to ((IndCount * CrossoverP) div 100)-1 do
    ind2[i].body := cross(ind[draw].body, ind[draw].body);

  for i := ((IndCount * CrossoverP) div 100) to IndCount - 1 do
  begin
    j := draw;
    ind2[i].body := ind[j].body;
    ind2[i].fitness := ind[j].fitness;
  end;
  if Elite then
  begin
    ind2[IndCount-1].body := ind[BestIndex].body;
    ind2[IndCount-1].fitness := ind[BestIndex].fitness;
  end
  else
  begin
    j := draw;
    ind2[IndCount-1].body := ind[j].body;
    ind2[IndCount-1].fitness := ind[j].fitness;
  end;
  for i := 0 to IndCount-1{((IndCount * CrossoverP) div 100)-1 }do
  begin
    ind[i].body := ind2[i].body;
    ind[i].fitness := ind2[i].fitness;
  end;
end;

{ Stop running started by Execute method }
procedure TGPopulation.Stop;
     { Internal declarations for method }
     { type }
     { . . . }
     { var }
     { . . . }
begin
 Running:=false;
end;

{ Perform crossover on to s-expressions }
function TGPopulation.cross(s1:string;s2:string):string;
var
  i, l1, l2, count, k1, k2: integer;
  flag, flag2: boolean;
  c1, c2: string;
begin
  l1 := length(s1);
  l2 := length(s2);
  repeat
    c2 := '';
    c1 := '';
    while c1 = '' do
    begin
      i := round(random(l1)) + 1;
      flag := false;
      c1 := ''; count := -1000;
      flag2 := true;
      while (i <= l1) and flag2 do
      begin
        if (s1[i] = '(') and (not flag) then
        begin flag := true; count := 0; end;
        if (s1[i] = '(') and flag then
          count := count + 1;
        if flag and ((count > 1)
          or ((s1[i] <> '(')
          and (s1[i] <> ')'))) then
          c1 := c1 + s1[i];
        if (s1[i] = ')') and flag then
          count := count - 1;
        if (s1[i] = ')') and flag and (count = 0) then
        begin flag := false; flag2 := false; end;
        i := i + 1;
      end;
    end;
    k1 := 0;
    while (k1 = 0) do
    begin
      flag := false;
      c2 := '';
      flag2 := true;
      i := round(random(l2) / 2) + 1;
      k1 := 0; count := -1000;
      k2 := l2 + 1;
      while (i <= l2) and flag2 do
      begin
        if (s2[i] = '(') and (not flag) then
        begin flag := true; count := 0; k1 := i; end;
        if (s2[i] = '(') and flag then
          count := count + 1;
        if (s2[i] = ')') and flag then
          count := count - 1;
        if (s2[i] = ')') and flag and (count = 0) then
        begin flag := false; flag2 := false; k2 := i; end;
        i := i + 1;
      end;
    end;
    c2 := copy(s2, 1, k1) + c1 + copy(s2, k2, l2 - k2 + 1);
  until length(c2) <= IndLength;
  cross := c2;
end;

function TGPopulation.draw:integer;
var
  r: cardinal;
  k: integer;
begin
  r := round(random(roulete[IndCount-1])) + 1;
  k := 0;
  while (r > roulete[k]) and (k < IndCount-1) do
  begin
    k := k + 1;
  end;
  draw := k;
end;


end.
